@@ -1,3 +1,19 @@
+#
+# Copyright 2007 David Snopek <dsnopek@gmail.com>
+#
+# This program is free software: you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation, either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program.  If not, see <http://www.gnu.org/licenses/>.
+#
 
 package POE::Component::MessageQueue::Client;
 
@@ -34,6 +50,13 @@ sub new
 	return $self;
 }
 
+sub get_subscribed_queue_names
+{
+	my $self = shift;
+	my @queues = @{ $self->{queue_names} || [] };
+	return @queues;
+}
+
 sub _add_queue_name
 {
 	my ($self, $queue_name) = @_;
