@@ -1,3 +1,19 @@
+#
+# Copyright 2007 David Snopek <dsnopek@gmail.com>
+#
+# This program is free software: you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation, either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program.  If not, see <http://www.gnu.org/licenses/>.
+#
 
 package POE::Component::MessageQueue::Storage;
 
@@ -17,6 +33,7 @@ sub new
 		message_stored    => undef,
 		dispatch_message  => undef,
 		destination_ready => undef,
+		shutdown_complete => undef,
 		# TODO: do something with this.
 		started           => 0
 	};
@@ -52,6 +69,13 @@ sub set_destination_ready_handler
 	undef;
 }
 
+sub set_shutdown_complete_handler
+{
+	my ($self, $handler) = @_;
+	$self->{shutdown_complete} = $handler;
+	undef;
+}
+
 sub set_logger
 {
 	my ($self, $logger) = @_;
@@ -125,6 +149,21 @@ sub disown
 	die "Abstract.";
 }
 
+# a semi-hidden alias that we need when using a storage engine
+# behind POE::Component::MessageQueue::Storage::Generic!
+sub storage_shutdown
+{
+	my $self = shift;
+	$self->shutdown();
+}
+
+sub shutdown
+{
+	my $self = shift;
+
+	die "Abstract.";
+}
+
 1;
 
 __END__
@@ -151,12 +190,18 @@ Takes a CODEREF which will get called back when a message has been successfully
 
 Takes a CODEREF which will get called back when a message has been retrieved from the store.  This will be called with three arguments: the message, the destination string, and the client id.  If no message could be retrieved the function will still be called but with the message undefined.
 
-=item set_destination_ready_header I<CODEREF>
+=item set_destination_ready_handler I<CODEREF>
 
 Takes a CODEREF which will get called back when a destination is ready to be claimed from again.  This is necessary for storage engines that will lock a destination while attempting to retrieve a message.  This handler will be called when the destination is unlocked so that message queue knows that it can claim more messages.  If your storage engine doesn't lock anything, you B<must> call this handler immediately after called the above handler.  
 
 It will be called with a single argument: the destination string.
 
+=item set_shutdown_complete_handler I<CODEREF>
+
+Takes a CODEREF which will get called when the storage engine has finished shutting down.  The shutdown process is started by calling I<shutdown()> on the storage engine (see below).
+
+It will be called without any arguments.
+
 =item set_logger I<SCALAR>
 
 Takes an object of type L<POE::Component::MessageQueue::Logger> that should be used for logging.
@@ -181,6 +226,10 @@ Takes the destination string and client id (or a HASHREF with keys "destination"
 
 Takes a destination and client id.  All messages which are owned by this client id on this destination should be marked as owned by nobody.
 
+=item shutdown
+
+Will start shutting down the storage engine.  The I<shutdown_complete> handler will be called when the storage engine has finished shutting down.  This should be a graceful operation.  Ie., The storage engine will attempt to clean-up and push messages to persistent storage if possible before calling the I<shutdown_complete> handler.
+
 =back
 
 =head1 SEE ALSO
