@@ -1,3 +1,19 @@
+#
+# Copyright 2007 David Snopek <dsnopek@gmail.com>
+#
+# This program is free software: you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation, either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program.  If not, see <http://www.gnu.org/licenses/>.
+#
 
 package POE::Component::MessageQueue::Storage::Throttled;
 use base qw(POE::Component::MessageQueue::Storage);
@@ -35,6 +51,9 @@ sub new
 	$self->{throttle_max}    = $throttle_max;
 	$self->{throttle_count}  = 0;
 
+	# a flag for shutting down
+	$self->{shutdown} = 0;
+
 	# we have to intercept the message_stored handler.
 	$self->{storage}->set_message_stored_handler(sub { return $self->_message_stored(@_); });
 
@@ -47,19 +66,21 @@ sub new
 sub set_dispatch_message_handler
 {
 	my ($self, $handler) = @_;
-	# We never need to call this directly, storage will!
-	#$self->SUPER::set_dispatch_message_handler( $handler );
 	$self->{storage}->set_dispatch_message_handler( $handler );
 }
 
 sub set_destination_ready_handler
 {
 	my ($self, $handler) = @_;
-	# We never need to call this directly, storage will!
-	#$self->SUPER::set_destination_ready_handler( $handler );
 	$self->{storage}->set_destination_ready_handler( $handler );
 }
 
+sub set_shutdown_complete_handler
+{
+	my ($self, $handler) = @_;
+	$self->{storage}->set_shutdown_complete_handler( $handler );
+}
+
 sub set_logger
 {
 	my ($self, $logger) = @_;
@@ -102,6 +123,7 @@ sub _throttle_remove
 	if ( exists $self->{throttle_buffer}->{$message_id} )
 	{
 		delete $self->{throttle_buffer}->{$message_id};
+		
 		return 1;
 	}
 
@@ -137,6 +159,13 @@ sub _message_stored
 	{
 		$self->{message_stored}->( $destination );
 	}
+
+	# if we are shutting down and there are no more message throttled, then
+	# we shutdown the underlying engine.
+	if ( $self->{shutdown} and $self->{throttle_count} == 0 )
+	{
+		$self->{storage}->shutdown();
+	}
 }
 
 sub get_next_message_id
@@ -200,6 +229,24 @@ sub disown
 	return shift->{storage}->disown(@_);
 }
 
+sub shutdown
+{
+	my $self = shift;
+
+	# we mark that we are shutting down.
+	$self->{shutdown} = 1;
+
+	if ( $self->{throttle_count} == 0 )
+	{
+		# if there are no throttled messages, then we can just start
+		# shutting down the underlying storage engine.
+		$self->{storage}->shutdown();
+	}
+
+	# otherwise, we will wait until we have none throttled, then we
+	# will call the underlying shutdown.
+}
+
 1;
 
 __END__
