@@ -5,7 +5,9 @@ CREATE TABLE messages
 	destination varchar(255) not null,
 	persistent  char(1) default 'Y' not null,
 	in_use_by   int,
-	body        text
+	body        text,
+	timestamp   int,
+	size        int
 );
 
 -- Improves performance some bit:
